@@ -1,8 +1,8 @@
 /*************************************************************************
 ** Color.cpp                                                            **
 **                                                                      **
-** This file is part of dvisvgm -- the DVI to SVG converter             **
-** Copyright (C) 2005-2015 Martin Gieseking <martin.gieseking@uos.de>   **
+** This file is part of dvisvgm -- a fast DVI to SVG converter          **
+** Copyright (C) 2005-2019 Martin Gieseking <martin.gieseking@uos.de>   **
 **                                                                      **
 ** This program is free software; you can redistribute it and/or        **
 ** modify it under the terms of the GNU General Public License as       **
@@ -18,8 +18,7 @@
 ** along with this program; if not, see <http://www.gnu.org/licenses/>. **
 *************************************************************************/
 
-#define _USE_MATH_DEFINES
-#include <config.h>
+#include <array>
 #include <algorithm>
 #include <cctype>
 #include <cmath>
@@ -27,38 +26,27 @@
 #include <cstring>
 #include <iomanip>
 #include <sstream>
-#include "Color.h"
+#include "Color.hpp"
+#include "utility.hpp"
 
 using namespace std;
 
 bool Color::SUPPRESS_COLOR_NAMES = true;
 
-const Color Color::BLACK(UInt32(0));
-const Color Color::WHITE(UInt8(255), UInt8(255), UInt8(255));
-const Color Color::TRANSPARENT(UInt32(0xff000000));
+const Color Color::BLACK(uint32_t(0));
+const Color Color::WHITE(uint8_t(255), uint8_t(255), uint8_t(255));
+const Color Color::TRANSPARENT(uint32_t(0xff000000));
 
 
-static inline UInt8 double_to_byte (double v) {
+static inline uint8_t double_to_byte (double v) {
 	v = max(0.0, min(1.0, v));
-	return UInt8(floor(255*v+0.5));
-}
-
-
-static void tolower (string &str) {
-	for (size_t i=0; i < str.length(); i++)
-		str[i] = tolower(str[i]);
-}
-
-
-Color::Color (const char *psname) {
-	if (!setPSName(psname, false))
-		setGray(UInt8(0));
+	return uint8_t(floor(255*v+0.5));
 }
 
 
 Color::Color (const string &psname) {
 	if (!setPSName(psname, false))
-		setGray(UInt8(0));
+		setGray(uint8_t(0));
 }
 
 
@@ -74,17 +62,18 @@ void Color::setRGB (double r, double g, double b) {
 bool Color::setPSName (string name, bool case_sensitive) {
 	if (name[0] == '#') {
 		char *p=0;
-		_rgb = UInt32(strtol(name.c_str()+1, &p, 16));
+		_rgb = uint32_t(strtol(name.c_str()+1, &p, 16));
 		while (isspace(*p))
 			p++;
 		return (*p == 0 && _rgb <= 0xFFFFFF);
 	}
-	// converted color constants from color.pro
-	static const struct ColorConstant {
+
+	struct ColorConstant {
 		const char *name;
-		const UInt32 rgb;
-	}
-	constants[] = {
+		const uint32_t rgb;
+	};
+	// converted color constants from color.pro
+	static const array<ColorConstant, 68> constants {{
 		{"Apricot",        0xFFAD7A},
 		{"Aquamarine",     0x2DFFB2},
 		{"Bittersweet",    0xC10200},
@@ -153,31 +142,26 @@ bool Color::setPSName (string name, bool case_sensitive) {
 		{"Yellow",         0xFFFF00},
 		{"YellowGreen",    0x8EFF42},
 		{"YellowOrange",   0xFF9300},
-	};
-	if (!case_sensitive) {
-		tolower(name);
-		for (size_t i=0; i < sizeof(constants)/sizeof(ColorConstant); i++) {
-			string cmpname = constants[i].name;
-			tolower(cmpname);
-			if (name == cmpname) {
-				_rgb = constants[i].rgb;
-				return true;
+	}};
+	if (case_sensitive) {
+		const ColorConstant cmppair = {name.c_str(), 0};
+		auto it = lower_bound(constants.begin(), constants.end(), cmppair,
+			[](const ColorConstant &c1, const ColorConstant &c2) {
+				return strcmp(c1.name, c2.name) < 0;
 			}
+		);
+		if (it != constants.end() && it->name == name) {
+			_rgb = it->rgb;
+			return true;
 		}
-		return false;
 	}
-
-	// binary search
-	int first=0, last=sizeof(constants)/sizeof(ColorConstant)-1;
-	while (first <= last) {
-		int mid = first+(last-first)/2;
-		int cmp = strcmp(constants[mid].name, name.c_str());
-		if (cmp > 0)
-			last = mid-1;
-		else if (cmp < 0)
-			first = mid+1;
-		else {
-			_rgb = constants[mid].rgb;
+	else {
+		name = util::tolower(name);
+		auto it = find_if(constants.begin(), constants.end(), [&](const ColorConstant &cc) {
+			return name == util::tolower(cc.name);
+		});
+		if (it != constants.end()) {
+			_rgb = it->rgb;
 			return true;
 		}
 	}
@@ -215,21 +199,26 @@ void Color::setCMYK (const std::valarray<double> &cmyk) {
 
 void Color::set (ColorSpace colorSpace, VectorIterator<double> &it) {
 	switch (colorSpace) {
-		case GRAY_SPACE: setGray(*it++); break;
-		case RGB_SPACE : setRGB(*it, *(it+1), *(it+2)); it+=3; break;
-		case LAB_SPACE : setLab(*it, *(it+1), *(it+2)); it+=3; break;
-		case CMYK_SPACE: setCMYK(*it, *(it+1), *(it+2), *(it+3)); it+=4; break;
+		case ColorSpace::GRAY: setGray(*it++); break;
+		case ColorSpace::RGB : setRGB(*it, *(it+1), *(it+2)); it+=3; break;
+		case ColorSpace::LAB : setLab(*it, *(it+1), *(it+2)); it+=3; break;
+		case ColorSpace::CMYK: setCMYK(*it, *(it+1), *(it+2), *(it+3)); it+=4; break;
 	}
 }
 
 
-void Color::operator *= (double c) {
-	UInt32 rgb=0;
-	for (int i=0; i < 3; i++) {
-		rgb |= UInt32(floor((_rgb & 0xff)*c+0.5)) << (8*i);
-		_rgb >>= 8;
+Color Color::operator *= (double c) {
+	if (abs(c) < 0.001)
+		_rgb &= 0xff000000;
+	else if (abs(c-trunc(c)) < 0.999) {
+		uint32_t rgb=0;
+		for (int i=0; i < 3; i++) {
+			rgb |= uint32_t((_rgb & 0xff)*c+0.5) << (8*i);
+			_rgb >>= 8;
+		}
+		_rgb = rgb;
 	}
-	_rgb = rgb;
+	return *this;
 }
 
 
@@ -250,10 +239,11 @@ string Color::rgbString () const {
  *  define a name for the current color. */
 string Color::svgColorString (bool rgbonly) const {
 	if (!rgbonly) {
-		static const struct ColorName {
-			UInt32 rgb;
+		struct ColorName {
+			uint32_t rgb;
 			const char *name;
-		} colornames[] = {
+		};
+		static const array<ColorName, 138> colornames {{
 			{0x000000, "black"},
 			{0x000080, "navy"},
 			{0x00008b, "darkblue"},
@@ -392,18 +382,13 @@ string Color::svgColorString (bool rgbonly) const {
 			{0xffffe0, "lightyellow"},
 			{0xfffff0, "ivory"},
 			{0xffffff, "white"}
-		};
-		int left=0;
-		int right=sizeof(colornames)/sizeof(ColorName)-1;
-		while (left <= right) {
-			int mid = left+(right-left)/2;
-			if (colornames[mid].rgb == _rgb)
-				return colornames[mid].name;
-			if (colornames[mid].rgb > _rgb)
-				right = mid-1;
-			else
-				left = mid+1;
-		}
+		}};
+		ColorName cmppair = {_rgb, 0};
+		auto it = lower_bound(colornames.begin(), colornames.end(), cmppair, [](const ColorName &c1, const ColorName &c2) {
+			return c1.rgb < c2.rgb;
+		});
+		if (it != colornames.end() && it->rgb == _rgb)
+			return it->name;
 	}
 	return rgbString();
 }
@@ -569,10 +554,9 @@ void Color::getLab (std::valarray<double> &lab) const {
 }
 
 
-static inline double sqr (double x)  {return x*x;}
+// static inline double sqr (double x)  {return x*x;}
 static inline double cube (double x) {return x*x*x;}
 
-
 void Color::Lab2XYZ (const valarray<double> &lab, valarray<double> &xyz) {
 	xyz.resize(3);
 	double wx=0.95047, wy=1.00, wz=1.08883;  // reference white D65
@@ -628,52 +612,59 @@ void Color::RGB2Lab (const valarray<double> &rgb, valarray<double> &lab) {
 }
 
 
+#if 0
 /** Returns the Delta E difference (CIE 2000) between this and another color. */
 double Color::deltaE (const Color &c) const {
 	double l1, a1, b1;
 	double l2, a2, b2;
 	getLab(l1, a1, b1);
 	c.getLab(l2, a2, b2);
-	double dl = l2-l1;
-	double lm = (l1+l2)/2;
-	double c1 = sqrt(a1*a1 + b1*b1);
-	double c2 = sqrt(a2*a2 + b2*b2);
-	double cm = (c1+c2)/2;
-	double g  = (1-sqrt(pow(cm, 7)/(pow(cm, 7)+pow(25.0, 7))))/2;
-	double aa1 = a1*(1+g);
-	double aa2 = a2*(1+g);
-	double cc1 = sqrt(aa1*aa1 + b1*b1);
-	double cc2 = sqrt(aa2*aa2 + b2*b2);
-	double ccm = (cc1+cc2)/2;
-	double dcc = cc2-cc1;
-	double h1  = atan2(b1, aa1)*180/M_PI;
-	if (h1 < 0) h1 += 360;
-	double h2  = atan2(b2, aa2)*180/M_PI;
-	if (h2 < 0)	h2 += 360;
-	double hm = (abs(h1-h2) > 180 ? (h1+h2+360) : (h1+h2))/2;
-	double t  = 1 - 0.17*cos(hm-30) + 0.24*cos(2*hm) + 0.32*cos(3*hm+6) - 0.2*cos(4*hm-63);
-	double dh = h2-h1;
-	if (h2-h1 < -180)
-		dh += 360;
-	else if (h2-h1 > 180)
-		dh -= 360;
-	double dhh = 2*sqrt(cc1*cc2)*sin(dh/2);
-	double sl = 1 + 0.015*(lm-50.0)*(lm-50.0)/sqrt(20.0+(lm-50.0));
-	double sc = 1 + 0.045*ccm;
-	double sh = 1 + 0.015*ccm*t;
-	double dtheta = 30*exp(-sqr(hm-275)/25);
-	double rc = 2*sqrt(pow(ccm, 7)/(pow(ccm, 7)+pow(25.0, 7)));
-	double rt = -rc*sin(2*dtheta);
-	return sqrt(sqr(dl/sl) + sqr(dcc/sc) + sqr(dhh/sh) + rt*dcc/sc*dhh/sh);
+	const double deltaL = l2-l1;
+	const double lBar = (l1+l2)/2;
+	const double c1 = sqrt(a1*a1 + b1*b1);
+	const double c2 = sqrt(a2*a2 + b2*b2);
+	const double cBar = (c1+c2)/2.0;
+	const double g = (1.0-sqrt(pow(cBar, 7.0)/(pow(cBar, 7.0)+6103515625.0)))/2.0;
+	const double aa1 = a1*(1.0+g);
+	const double aa2 = a2*(1.0+g);
+	const double cc1 = sqrt(aa1*aa1 + b1*b1);
+	const double cc2 = sqrt(aa2*aa2 + b2*b2);
+	const double ccBar = (cc1+cc2)/2.0;
+	const double deltaCC = cc2-cc1;
+	double hh1 = atan2(b1, aa1);
+	if (hh1 < 0) hh1 += deg2rad(360);
+	double hh2 = atan2(b2, aa2);
+	if (hh2 < 0) hh2 += deg2rad(360);
+	const double hhBar = (std::abs(hh1-hh2) > deg2rad(180) ? (hh1+hh2+deg2rad(360)) : (hh1+hh2))/2.0;
+	const double t = 1.0
+		- 0.17*cos(hhBar-deg2rad(30.0))
+		+ 0.24*cos(2.0*hhBar)
+		+ 0.32*cos(3.0*hhBar+deg2rad(6.0))
+		- 0.20*cos(4.0*hhBar-deg2rad(63.0));
+	double deltaHH = hh2-hh1;
+	if (deltaHH < deg2rad(-180))
+		deltaHH += deg2rad(360);
+	else if (deltaHH > deg2rad(180))
+		deltaHH -= deg2rad(360);
+	const double deltaHHH = 2.0*sqrt(cc1*cc2)*sin(deltaHH/2.0);
+	const double sl = 1.0 + 0.015*sqr(lBar-50.0)/sqrt(20.0+sqr(lBar-50.0));
+	const double sc = 1.0 + 0.045*ccBar;
+	const double sh = 1.0 + 0.015*ccBar*t;
+	const double deltaTheta = deg2rad(30)*exp(-sqr((hhBar-deg2rad(275))/deg2rad(25)));
+	const double rc = 2.0*sqrt(pow(ccBar, 7)/(pow(ccBar, 7)+6103515625.0));
+	const double rt = -rc*sin(2.0*deltaTheta);
+	const double deltaE = sqrt(sqr(deltaL/sl) + sqr(deltaCC/sc) + sqr(deltaHHH/sh) + rt*deltaCC/sc*deltaHHH/sh);
+	return deltaE;
 }
+#endif
 
 
 int Color::numComponents (ColorSpace colorSpace) {
 	switch (colorSpace) {
-		case GRAY_SPACE: return 1;
-		case LAB_SPACE:
-		case RGB_SPACE:  return 3;
-		case CMYK_SPACE: return 4;
+		case ColorSpace::GRAY: return 1;
+		case ColorSpace::LAB:
+		case ColorSpace::RGB:  return 3;
+		case ColorSpace::CMYK: return 4;
 	}
 	return 0;
 }
