@@ -2,7 +2,7 @@
 ** PsSpecialHandler.cpp                                                 **
 **                                                                      **
 ** This file is part of dvisvgm -- a fast DVI to SVG converter          **
-** Copyright (C) 2005-2018 Martin Gieseking <martin.gieseking@uos.de>   **
+** Copyright (C) 2005-2019 Martin Gieseking <martin.gieseking@uos.de>   **
 **                                                                      **
 ** This program is free software; you can redistribute it and/or        **
 ** modify it under the terms of the GNU General Public License as       **
@@ -18,12 +18,14 @@
 ** along with this program; if not, see <http://www.gnu.org/licenses/>. **
 *************************************************************************/
 
+#include <array>
 #include <cmath>
 #include <fstream>
 #include <memory>
 #include <sstream>
 #include "EPSFile.hpp"
 #include "FileFinder.hpp"
+#include "FileSystem.hpp"
 #include "Message.hpp"
 #include "PathClipper.hpp"
 #include "PSPattern.hpp"
@@ -34,25 +36,16 @@
 #include "TensorProductPatch.hpp"
 #include "TriangularPatch.hpp"
 
-
 using namespace std;
 
 
-static inline double str2double (const string &str) {
-	double ret;
-	istringstream iss(str);
-	iss >> ret;
-	return ret;
-}
-
-
 bool PsSpecialHandler::COMPUTE_CLIPPATHS_INTERSECTIONS = false;
 bool PsSpecialHandler::SHADING_SEGMENT_OVERLAP = false;
 int PsSpecialHandler::SHADING_SEGMENT_SIZE = 20;
 double PsSpecialHandler::SHADING_SIMPLIFY_DELTA = 0.01;
 
 
-PsSpecialHandler::PsSpecialHandler () : _psi(this), _actions(), _previewFilter(_psi), _psSection(PS_NONE), _xmlnode(), _savenode()
+PsSpecialHandler::PsSpecialHandler () : _psi(this), _actions(), _previewFilter(_psi), _xmlnode(), _savenode()
 {
 }
 
@@ -69,7 +62,7 @@ void PsSpecialHandler::initialize () {
 	if (_psSection == PS_NONE) {
 		initgraphics();
 		// execute dvips prologue/header files
-		for (const char *fname : {"tex.pro", "texps.pro", "special.pro"})
+		for (const char *fname : {"tex.pro", "texps.pro", "special.pro", "color.pro"})
 			processHeaderFile(fname);
 		// disable bop/eop operators to prevent side-effects by
 		// unexpected bops/eops present in PS specials
@@ -85,7 +78,8 @@ void PsSpecialHandler::initgraphics () {
 	_linecap = _linejoin = 0;  // butt end caps and miter joins
 	_miterlimit = 4;
 	_xmlnode = _savenode = nullptr;
-	_opacityalpha = 1;  // fully opaque
+	_opacityalpha = _shapealpha = 1;  // fully opaque
+	_blendmode = 0; // "normal" mode (no blending)
 	_sx = _sy = _cos = 1.0;
 	_pattern = nullptr;
 	_currentcolor = Color::BLACK;
@@ -170,17 +164,17 @@ void PsSpecialHandler::executeAndSync (istream &is, bool updatePos) {
 }
 
 
-void PsSpecialHandler::preprocess (const char *prefix, istream &is, SpecialActions &actions) {
+void PsSpecialHandler::preprocess (const string &prefix, istream &is, SpecialActions &actions) {
 	initialize();
 	if (_psSection != PS_HEADERS)
 		return;
 
 	_actions = &actions;
-	if (*prefix == '!') {
+	if (prefix == "!") {
 		_headerCode += "\n";
 		_headerCode += string(istreambuf_iterator<char>(is), istreambuf_iterator<char>());
 	}
-	else if (strcmp(prefix, "header=") == 0) {
+	else if (prefix == "header=") {
 		// read and execute PS header file
 		string fname;
 		is >> fname;
@@ -189,9 +183,9 @@ void PsSpecialHandler::preprocess (const char *prefix, istream &is, SpecialActio
 }
 
 
-bool PsSpecialHandler::process (const char *prefix, istream &is, SpecialActions &actions) {
+bool PsSpecialHandler::process (const string &prefix, istream &is, SpecialActions &actions) {
 	// process PS headers only once (in prescan)
-	if (*prefix == '!' || strcmp(prefix, "header=") == 0)
+	if (prefix == "!" || prefix == "header=")
 		return true;
 
 	_actions = &actions;
@@ -199,23 +193,23 @@ bool PsSpecialHandler::process (const char *prefix, istream &is, SpecialActions
 	if (_psSection != PS_BODY)
 		enterBodySection();
 
-	if (*prefix == '"' || strcmp(prefix, "pst:") == 0) {
+	if (prefix == "\"" || prefix == "pst:") {
 		// read and execute literal PostScript code (isolated by a wrapping save/restore pair)
 		moveToDVIPos();
 		_psi.execute("\n@beginspecial @setspecial ");
 		executeAndSync(is, false);
 		_psi.execute("\n@endspecial ");
 	}
-	else if (strcmp(prefix, "psfile=") == 0 || strcmp(prefix, "PSfile=") == 0) {
+	else if (prefix == "psfile=" || prefix == "PSfile=" || prefix == "pdffile=") {
 		if (_actions) {
 			StreamInputReader in(is);
-			string fname = in.getQuotedString(in.peek() == '"' ? '"' : 0);
+			const string fname = in.getQuotedString(in.peek() == '"' ? '"' : 0);
 			unordered_map<string,string> attr;
 			in.parseAttributes(attr);
-			psfile(fname, attr);
+			imgfile(prefix == "pdffile=" ? FileType::PDF : FileType::EPS, fname, attr);
 		}
 	}
-	else if (strcmp(prefix, "ps::") == 0) {
+	else if (prefix == "ps::") {
 		if (_actions)
 			_actions->finishLine();  // reset DVI position on next DVI command
 		if (is.peek() == '[') {
@@ -249,7 +243,7 @@ bool PsSpecialHandler::process (const char *prefix, istream &is, SpecialActions
 		StreamInputReader in(is);
 		if (in.check(" plotfile ")) { // ps: plotfile fname
 			string fname = in.getString();
-			ifstream ifs(fname.c_str());
+			ifstream ifs(fname);
 			if (ifs)
 				_psi.execute(ifs);
 			else
@@ -266,46 +260,63 @@ bool PsSpecialHandler::process (const char *prefix, istream &is, SpecialActions
 }
 
 
-/** Handles psfile special.
- *  @param[in] fname EPS file to be included
- *  @param[in] attr attributes given with \\special psfile */
-void PsSpecialHandler::psfile (const string &fname, const unordered_map<string,string> &attr) {
+/** Handles a psfile/pdffile special which places an external EPS/PDF graphic
+ *  at the current DVI position. The lower left corner (llx,lly) of the
+ *  given bounding box is placed at the DVI position.
+ *  @param[in] filetype type of file to process (EPS or PDF)
+ *  @param[in] fname EPS/PDF file to be included
+ *  @param[in] attr attributes given with psfile/pdffile special */
+void PsSpecialHandler::imgfile (FileType filetype, const string &fname, const unordered_map<string,string> &attr) {
+	// prevent warning about missing image file "/dev/null" which is
+	// added by option "psfixbb" of the preview package
+	if (fname == "/dev/null")
+		return;
+
 	const char *filepath = FileFinder::instance().lookup(fname, false);
+	if (!filepath && FileSystem::exists(fname))
+		filepath = fname.c_str();
 	if (!filepath) {
-		Message::wstream(true) << "file '" << fname << "' not found in special 'psfile'\n";
+		Message::wstream(true) << "file '" << fname << "' not found\n";
 		return;
 	}
 	unordered_map<string,string>::const_iterator it;
 
-	// bounding box of EPS figure (lower left and upper right corner)
-	double llx = (it = attr.find("llx")) != attr.end() ? str2double(it->second) : 0;
-	double lly = (it = attr.find("lly")) != attr.end() ? str2double(it->second) : 0;
-	double urx = (it = attr.find("urx")) != attr.end() ? str2double(it->second) : 0;
-	double ury = (it = attr.find("ury")) != attr.end() ? str2double(it->second) : 0;
+	// bounding box of EPS figure in PS point units (lower left and upper right corner)
+	double llx = (it = attr.find("llx")) != attr.end() ? stod(it->second) : 0;
+	double lly = (it = attr.find("lly")) != attr.end() ? stod(it->second) : 0;
+	double urx = (it = attr.find("urx")) != attr.end() ? stod(it->second) : 0;
+	double ury = (it = attr.find("ury")) != attr.end() ? stod(it->second) : 0;
+	int pageno = (it = attr.find("page")) != attr.end() ? stoi(it->second, nullptr, 10) : 1;
 
-	// desired width/height of resulting figure
-	double rwi = (it = attr.find("rwi")) != attr.end() ? str2double(it->second)/10.0 : -1;
-	double rhi = (it = attr.find("rhi")) != attr.end() ? str2double(it->second)/10.0 : -1;
+	if (filetype == FileType::PDF && llx == 0 && lly == 0 && urx == 0 && ury == 0) {
+		BoundingBox pagebox = _psi.pdfPageBox(fname, pageno);
+		if (pagebox.valid()) {
+			llx = pagebox.minX();
+			lly = pagebox.minY();
+			urx = pagebox.maxX();
+			ury = pagebox.maxY();
+		}
+	}
+
+	// desired width and height of the untransformed figure in PS point units
+	double rwi = (it = attr.find("rwi")) != attr.end() ? stod(it->second)/10.0 : -1;
+	double rhi = (it = attr.find("rhi")) != attr.end() ? stod(it->second)/10.0 : -1;
 	if (rwi == 0 || rhi == 0 || urx-llx == 0 || ury-lly == 0)
 		return;
 
 	// user transformations (default values chosen according to dvips manual)
-	double hoffset = (it = attr.find("hoffset")) != attr.end() ? str2double(it->second) : 0;
-	double voffset = (it = attr.find("voffset")) != attr.end() ? str2double(it->second) : 0;
-//	double hsize   = (it = attr.find("hsize")) != attr.end() ? str2double(it->second) : 612;
-//	double vsize   = (it = attr.find("vsize")) != attr.end() ? str2double(it->second) : 792;
-	double hscale  = (it = attr.find("hscale")) != attr.end() ? str2double(it->second) : 100;
-	double vscale  = (it = attr.find("vscale")) != attr.end() ? str2double(it->second) : 100;
-	double angle   = (it = attr.find("angle")) != attr.end() ? str2double(it->second) : 0;
-
-	Matrix m(1);
-	m.rotate(angle).scale(hscale/100, vscale/100).translate(hoffset, voffset);
-	BoundingBox bbox(llx, lly, urx, ury);
-	bbox.transform(m);
+	// order of transformations: rotate, scale, translate/offset
+	double hoffset = (it = attr.find("hoffset")) != attr.end() ? stod(it->second) : 0;
+	double voffset = (it = attr.find("voffset")) != attr.end() ? stod(it->second) : 0;
+//	double hsize   = (it = attr.find("hsize")) != attr.end() ? stod(it->second) : 612;
+//	double vsize   = (it = attr.find("vsize")) != attr.end() ? stod(it->second) : 792;
+	double hscale  = (it = attr.find("hscale")) != attr.end() ? stod(it->second) : 100;
+	double vscale  = (it = attr.find("vscale")) != attr.end() ? stod(it->second) : 100;
+	double angle   = (it = attr.find("angle")) != attr.end() ? stod(it->second) : 0;
 
 	// compute factors to scale the bounding box to width rwi and height rhi
-	double sx = rwi/bbox.width();
-	double sy = rhi/bbox.height();
+	double sx = rwi/abs(llx-urx);
+	double sy = rhi/abs(lly-ury);
 	if (sx == 0 || sy == 0)
 		return;
 
@@ -314,34 +325,34 @@ void PsSpecialHandler::psfile (const string &fname, const unordered_map<string,s
 	if (sx < 0) sx = sy = 1.0;   // neither rwi nor rhi set
 
 	// save current DVI position
-	const double x = _actions->getX();
-	const double y = _actions->getY();
+	double x = _actions->getX();
+	double y = _actions->getY();
 
 	// all following drawings are relative to (0,0)
 	_actions->setX(0);
 	_actions->setY(0);
 	moveToDVIPos();
 
-	// transform current DVI position and bounding box location
-	// according to current transformation matrix
-	DPair llTrans = _actions->getMatrix()*DPair(llx, -lly);
-	DPair urTrans = _actions->getMatrix()*DPair(urx, -ury);
-	DPair dviposTrans = _actions->getMatrix()*DPair(x, y);
-
 	auto groupNode = util::make_unique<XMLElementNode>("g");  // append following elements to this group
 	_xmlnode = groupNode.get();
-	_psi.execute("\n@beginspecial @setspecial "); // enter \special environment
-	EPSFile epsfile(filepath);
-	_psi.limit(epsfile.pslength());  // limit the number of bytes going to be processed
-	_psi.execute(epsfile.istream()); // process EPS file
-	_psi.limit(0);                   // disable limitation
-	_psi.execute("\n@endspecial ");  // leave special environment
+	_psi.execute(
+		"\n@beginspecial @setspecial"          // enter special environment
+		"/setpagedevice{@setpagedevice}def"    // activate processing of operator "setpagedevice"
+		"[1 0 0 -1 0 0] setmatrix"             // don't apply outer PS transformations
+		"/FirstPage "+to_string(pageno)+" def" // set number of fisrt page to convert (PDF only)
+		"/LastPage "+to_string(pageno)+" def"  // set number of last page to convert (PDF only)
+		"(" + string(filepath) + ")run "       // execute file content
+		"@endspecial "                         // leave special environment
+	);
 	if (!groupNode->empty()) {       // has anything been drawn?
 		Matrix matrix(1);
-		matrix.rotate(angle).scale(hscale/100, vscale/100).translate(hoffset, voffset);
-		matrix.translate(-llTrans);
-		matrix.scale(sx, sy);          // resize image to width "rwi" and height "rhi"
-		matrix.translate(dviposTrans); // move image to current DVI position
+		if (filetype == FileType::PDF)
+			matrix.translate(-llx, -lly).scale(1, -1); //.translate(0, lly);  // flip vertically
+		else
+			matrix.translate(-llx, lly);
+		matrix.scale(sx, sy).rotate(-angle).scale(hscale/100, vscale/100);
+		matrix.translate(x+hoffset, y-voffset); // move image to current DVI position
+		matrix.rmultiply(_actions->getMatrix());
 		if (!matrix.isIdentity())
 			groupNode->addAttribute("transform", matrix.getSVG());
 		_actions->appendToPage(std::move(groupNode));
@@ -354,10 +365,12 @@ void PsSpecialHandler::psfile (const string &fname, const unordered_map<string,s
 	moveToDVIPos();
 
 	// update bounding box
-	m.scale(sx, -sy);
-	m.translate(dviposTrans);
-	bbox = BoundingBox(DPair(0, 0), abs(urTrans-llTrans));
-	bbox.transform(m);
+	BoundingBox bbox(0, 0, urx-llx, ury-lly);
+	Matrix matrix(1);
+	matrix.scale(sx, -sy).rotate(-angle).scale(hscale/100, vscale/100);
+	matrix.translate(x+hoffset, y-voffset);
+	matrix.rmultiply(_actions->getMatrix());
+	bbox.transform(matrix);
 	_actions->embed(bbox);
 }
 
@@ -393,38 +406,25 @@ static bool transform_box_extents (const Matrix &matrix, double &w, double &h, d
 void PsSpecialHandler::dviEndPage (unsigned, SpecialActions &actions) {
 	BoundingBox bbox;
 	if (_previewFilter.getBoundingBox(bbox)) {  // is there any data written by preview package?
-		double w = max(0.0, _previewFilter.width());
-		double h = max(0.0, _previewFilter.height());
-		double d = max(0.0, _previewFilter.depth());
+		double w=0, h=0, d=0;
 		if (actions.getBBoxFormatString() == "preview" || actions.getBBoxFormatString() == "min") {
 			if (actions.getBBoxFormatString() == "preview") {
+				w = max(0.0, _previewFilter.width());
+				h = max(0.0, _previewFilter.height());
+				d = max(0.0, _previewFilter.depth());
 				actions.bbox() = bbox;
 				Message::mstream() << "\napplying bounding box set by";
 			}
 			else {
-				// compute height, depth and width depending on the
-				// tight bounding box derived from the objects on the page
-				double y0 = bbox.maxY()-h;      // y coordinate of the baseline
-				h = actions.bbox().maxY()-y0;
-				if (h < 0) {
-					h = 0;
-					d = actions.bbox().height();
-				}
-				else {
-					d = y0-actions.bbox().minY();
-					if (d < 0) {
-						h = actions.bbox().height();
-						d = 0;
-					}
-				}
 				w = actions.bbox().width();
+				h = max(0.0, -actions.bbox().minY());
+				d = max(0.0, actions.bbox().maxY());
 				Message::mstream() << "\ncomputing extents based on data set by";
 			}
 			Message::mstream() << " preview package (version " << _previewFilter.version() << ")\n";
 
 			// apply page transformations to box extents
-			Matrix pagetrans;
-			actions.getPageTransform(pagetrans);
+			Matrix pagetrans = actions.getPageTransformation();
 			bool isBaselineHorizontal = transform_box_extents(pagetrans, w, h, d);
 			actions.bbox().lock();
 
@@ -437,28 +437,28 @@ void PsSpecialHandler::dviEndPage (unsigned, SpecialActions &actions) {
 					"height=" << XMLString(h*bp2pt) << "pt, "
 					"depth=" << XMLString(d*bp2pt) << "pt\n";
 			}
-		}
 #if 0
-		XMLElementNode *rect = new XMLElementNode("rect");
-		rect->addAttribute("x", actions.bbox().minX());
-		rect->addAttribute("y", actions.bbox().minY());
-		rect->addAttribute("width", w);
-		rect->addAttribute("height", h+d);
-		rect->addAttribute("fill", "none");
-		rect->addAttribute("stroke", "red");
-		rect->addAttribute("stroke-width", "0.5");
-		actions.appendToPage(rect);
-		if (d > 0) {
-			XMLElementNode *line = new XMLElementNode("line");
-			line->addAttribute("x1", actions.bbox().minX());
-			line->addAttribute("y1", actions.bbox().minY()+h);
-			line->addAttribute("x2", actions.bbox().maxX());
-			line->addAttribute("y2", actions.bbox().minY()+h);
-			line->addAttribute("stroke", "blue");
-			line->addAttribute("stroke-width", "0.5");
-			actions.appendToPage(line);
-		}
+			auto rect = util::make_unique<XMLElementNode>("rect");
+			rect->addAttribute("x", actions.bbox().minX());
+			rect->addAttribute("y", actions.bbox().minY());
+			rect->addAttribute("width", w);
+			rect->addAttribute("height", h+d);
+			rect->addAttribute("fill", "none");
+			rect->addAttribute("stroke", "red");
+			rect->addAttribute("stroke-width", "0.1");
+			actions.appendToPage(std::move(rect));
+			if (d > 0) {
+				auto line = util::make_unique<XMLElementNode>("line");
+				line->addAttribute("x1", actions.bbox().minX());
+				line->addAttribute("y1", actions.bbox().minY()+h);
+				line->addAttribute("x2", actions.bbox().maxX());
+				line->addAttribute("y2", actions.bbox().minY()+h);
+				line->addAttribute("stroke", "blue");
+				line->addAttribute("stroke-width", "0.1");
+				actions.appendToPage(std::move(line));
+			}
 #endif
+		}
 	}
 	// close dictionary TeXDict and execute end-hook if defined
 	if (_psSection == PS_BODY) {
@@ -470,6 +470,21 @@ void PsSpecialHandler::dviEndPage (unsigned, SpecialActions &actions) {
 
 ///////////////////////////////////////////////////////
 
+void PsSpecialHandler::setpagedevice (std::vector<double> &p) {
+	_linewidth = 1;
+	_linecap = _linejoin = 0;  // butt end caps and miter joins
+	_miterlimit = 4;
+	_opacityalpha = _shapealpha = 1;  // fully opaque
+	_blendmode = 0; // "normal" mode (no blending)
+	_sx = _sy = _cos = 1.0;
+	_pattern = nullptr;
+	_currentcolor = Color::BLACK;
+	_dashoffset = 0;
+	_dashpattern.clear();
+	_path.clear();
+}
+
+
 void PsSpecialHandler::gsave (vector<double>&) {
 	_clipStack.dup();
 }
@@ -515,11 +530,22 @@ void PsSpecialHandler::closepath (vector<double>&) {
 }
 
 
+static string css_blendmode_name (int mode) {
+	static const array<const char*,16> modenames = {{
+	  "normal",  "multiply",  "screen", "overlay", "soft-light", "hard-light", "color-dodge", "color-burn",
+	  "darken", "lighten", "difference", "exclusion", "hue", "saturation", "color", "luminosity"
+	}};
+	if (mode < 0 || mode > 15)
+		return "";
+	return modenames[mode];
+}
+
+
 /** Draws the current path recorded by previously executed path commands (moveto, lineto,...).
  *  @param[in] p not used */
 void PsSpecialHandler::stroke (vector<double> &p) {
 	_path.removeRedundantCommands();
-	if ((_path.empty() && !_clipStack.clippathLoaded()) || !_actions)
+	if ((_path.empty() && !_clipStack.prependedPath()) || !_actions)
 		return;
 
 	BoundingBox bbox;
@@ -528,8 +554,8 @@ void PsSpecialHandler::stroke (vector<double> &p) {
 		if (!_xmlnode)
 			bbox.transform(_actions->getMatrix());
 	}
-	if (_clipStack.clippathLoaded() && _clipStack.top())
-		_path.prepend(*_clipStack.top());
+	if (_clipStack.prependedPath())
+		_path.prepend(*_clipStack.prependedPath());
 	unique_ptr<XMLElementNode> path;
 	Pair<double> point;
 	if (_path.isDot(point)) {  // zero-length path?
@@ -564,8 +590,10 @@ void PsSpecialHandler::stroke (vector<double> &p) {
 			path->addAttribute("stroke-linecap", _linecap == 1 ? "round" : "square");
 		if (_linejoin > 0)    // default value is "miter", no need to set it explicitly
 			path->addAttribute("stroke-linejoin", _linecap == 1 ? "round" : "bevel");
-		if (_opacityalpha < 1)
-			path->addAttribute("stroke-opacity", _opacityalpha);
+		if (_opacityalpha < 1 || _shapealpha < 1)
+			path->addAttribute("stroke-opacity", _opacityalpha*_shapealpha);
+		if (_blendmode > 0 && _blendmode < 16)
+			path->addAttribute("style", "mix-blend-mode:"+css_blendmode_name(_blendmode));
 		if (!_dashpattern.empty()) {
 			ostringstream oss;
 			for (size_t i=0; i < _dashpattern.size(); i++) {
@@ -578,13 +606,13 @@ void PsSpecialHandler::stroke (vector<double> &p) {
 				path->addAttribute("stroke-dashoffset", _dashoffset);
 		}
 	}
-	if (path && _clipStack.top()) {
+	if (path && _clipStack.path()) {
 		// assign clipping path and clip bounding box
 		path->addAttribute("clip-path", XMLString("url(#clip")+XMLString(_clipStack.topID())+")");
 		BoundingBox clipbox;
-		_clipStack.top()->computeBBox(clipbox);
+		_clipStack.path()->computeBBox(clipbox);
 		bbox.intersect(clipbox);
-		_clipStack.setClippathLoaded(false);
+		_clipStack.removePrependedPath();
 	}
 	if (_xmlnode)
 		_xmlnode->append(std::move(path));
@@ -601,7 +629,7 @@ void PsSpecialHandler::stroke (vector<double> &p) {
  *  @param[in] evenodd true: use even-odd fill algorithm, false: use nonzero fill algorithm */
 void PsSpecialHandler::fill (vector<double> &p, bool evenodd) {
 	_path.removeRedundantCommands();
-	if ((_path.empty() && !_clipStack.clippathLoaded()) || !_actions)
+	if ((_path.empty() && !_clipStack.prependedPath()) || !_actions)
 		return;
 
 	// compute bounding box
@@ -612,8 +640,8 @@ void PsSpecialHandler::fill (vector<double> &p, bool evenodd) {
 		if (!_xmlnode)
 			bbox.transform(_actions->getMatrix());
 	}
-	if (_clipStack.clippathLoaded() && _clipStack.top())
-		_path.prepend(*_clipStack.top());
+	if (_clipStack.prependedPath())
+		_path.prepend(*_clipStack.prependedPath());
 
 	ostringstream oss;
 	_path.writeSVG(oss, SVGTree::RELATIVE_PATH_CMDS);
@@ -623,18 +651,20 @@ void PsSpecialHandler::fill (vector<double> &p, bool evenodd) {
 		path->addAttribute("fill", XMLString("url(#")+_pattern->svgID()+")");
 	else if (_actions->getColor() != Color::BLACK || _savenode)
 		path->addAttribute("fill", _actions->getColor().svgColorString());
-	if (_clipStack.top()) {
+	if (_clipStack.path()) {
 		// assign clipping path and clip bounding box
 		path->addAttribute("clip-path", XMLString("url(#clip")+XMLString(_clipStack.topID())+")");
 		BoundingBox clipbox;
-		_clipStack.top()->computeBBox(clipbox);
+		_clipStack.path()->computeBBox(clipbox);
 		bbox.intersect(clipbox);
-		_clipStack.setClippathLoaded(false);
+		_clipStack.removePrependedPath();
 	}
 	if (evenodd)  // SVG default fill rule is "nonzero" algorithm
 		path->addAttribute("fill-rule", "evenodd");
-	if (_opacityalpha < 1)
-		path->addAttribute("fill-opacity", _opacityalpha);
+	if (_opacityalpha < 1 || _shapealpha < 1)
+		path->addAttribute("fill-opacity", _opacityalpha*_shapealpha);
+	if (_blendmode > 0 && _blendmode < 16)
+		path->addAttribute("style", "mix-blend-mode:"+css_blendmode_name(_blendmode));
 	if (_xmlnode)
 		_xmlnode->append(std::move(path));
 	else {
@@ -645,7 +675,7 @@ void PsSpecialHandler::fill (vector<double> &p, bool evenodd) {
 }
 
 
-/** Creates a Matrix object out of a given sequence of 6 double values.
+/** Creates a Matrix object from a given sequence of 6 double values.
  *  The given values must be arranged in PostScript matrix order.
  *  @param[in] v vector containing the matrix values
  *  @param[in] startindex vector index of first component
@@ -749,10 +779,8 @@ void PsSpecialHandler::initclip (vector<double> &) {
 
 /** Assigns the current clipping path to the graphics path. */
 void PsSpecialHandler::clippath (std::vector<double>&) {
-	if (!_clipStack.empty()) {
-		_clipStack.setClippathLoaded(true);
-		_path.clear();
-	}
+	if (!_clipStack.empty())
+		_clipStack.setPrependedPath();
 }
 
 
@@ -772,7 +800,7 @@ void PsSpecialHandler::clip (vector<double>&, bool evenodd) {
  *  computed by intersecting the current one with the given path.
  *  @param[in] path path used to restrict the clipping region
  *  @param[in] evenodd true: use even-odd fill algorithm, false: use nonzero fill algorithm */
-void PsSpecialHandler::clip (Path &path, bool evenodd) {
+void PsSpecialHandler::clip (Path path, bool evenodd) {
 	// when this method is called, _path contains the clipping path
 	if (path.empty() || !_actions)
 		return;
@@ -782,37 +810,41 @@ void PsSpecialHandler::clip (Path &path, bool evenodd) {
 
 	if (!_actions->getMatrix().isIdentity())
 		path.transform(_actions->getMatrix());
+	if (_clipStack.prependedPath())
+		path.prepend(*_clipStack.prependedPath());
 
 	int oldID = _clipStack.topID();
 
 	ostringstream oss;
+	bool pathReplaced;
 	if (!COMPUTE_CLIPPATHS_INTERSECTIONS || oldID < 1) {
-		_clipStack.replace(path);
+		pathReplaced = _clipStack.replace(path);
 		path.writeSVG(oss, SVGTree::RELATIVE_PATH_CMDS);
 	}
 	else {
 		// compute the intersection of the current clipping path with the current graphics path
-		const Path *oldPath = _clipStack.top();
+		const Path *oldPath = _clipStack.path();
 		Path intersectedPath(windingRule);
 		PathClipper clipper;
 		clipper.intersect(*oldPath, path, intersectedPath);
-		_clipStack.replace(intersectedPath);
+		pathReplaced = _clipStack.replace(intersectedPath);
 		intersectedPath.writeSVG(oss, SVGTree::RELATIVE_PATH_CMDS);
 	}
+	if (pathReplaced) {
+		auto pathElem = util::make_unique<XMLElementNode>("path");
+		pathElem->addAttribute("d", oss.str());
+		if (evenodd)
+			pathElem->addAttribute("clip-rule", "evenodd");
 
-	auto pathElem = util::make_unique<XMLElementNode>("path");
-	pathElem->addAttribute("d", oss.str());
-	if (evenodd)
-		pathElem->addAttribute("clip-rule", "evenodd");
+		int newID = _clipStack.topID();
+		auto clipElem = util::make_unique<XMLElementNode>("clipPath");
+		clipElem->addAttribute("id", XMLString("clip")+XMLString(newID));
+		if (!COMPUTE_CLIPPATHS_INTERSECTIONS && oldID)
+			clipElem->addAttribute("clip-path", XMLString("url(#clip")+XMLString(oldID)+")");
 
-	int newID = _clipStack.topID();
-	auto clipElem = util::make_unique<XMLElementNode>("clipPath");
-	clipElem->addAttribute("id", XMLString("clip")+XMLString(newID));
-	if (!COMPUTE_CLIPPATHS_INTERSECTIONS && oldID)
-		clipElem->addAttribute("clip-path", XMLString("url(#clip")+XMLString(oldID)+")");
-
-	clipElem->append(std::move(pathElem));
-	_actions->appendToDefs(std::move(clipElem));
+		clipElem->append(std::move(pathElem));
+		_actions->appendToDefs(std::move(clipElem));
+	}
 }
 
 
@@ -1038,13 +1070,12 @@ void PsSpecialHandler::processLatticeTriangularPatchMesh (ColorSpace colorSpace,
 }
 
 
-/** Clears current path */
+/** Clears current path. */
 void PsSpecialHandler::newpath (vector<double> &p) {
-	bool drawing = (p[0] > 0);
-	if (!drawing || !_clipStack.clippathLoaded()) {
-		_path.clear();
-		_clipStack.setClippathLoaded(false);
-	}
+	bool calledByNewpathOp = (p[0] > 0);
+	if (calledByNewpathOp)  // function triggered by PS operator 'newpath'?
+		_clipStack.removePrependedPath();
+	_path.clear();
 }
 
 
@@ -1093,7 +1124,7 @@ void PsSpecialHandler::rotate (vector<double> &p) {
 
 
 void PsSpecialHandler::setgray (vector<double> &p) {
-	_pattern = 0;
+	_pattern = nullptr;
 	_currentcolor.setGray(p[0]);
 	if (_actions)
 		_actions->setColor(_currentcolor);
@@ -1101,7 +1132,7 @@ void PsSpecialHandler::setgray (vector<double> &p) {
 
 
 void PsSpecialHandler::setrgbcolor (vector<double> &p) {
-	_pattern= 0;
+	_pattern= nullptr;
 	_currentcolor.setRGB(p[0], p[1], p[2]);
 	if (_actions)
 		_actions->setColor(_currentcolor);
@@ -1109,7 +1140,7 @@ void PsSpecialHandler::setrgbcolor (vector<double> &p) {
 
 
 void PsSpecialHandler::setcmykcolor (vector<double> &p) {
-	_pattern = 0;
+	_pattern = nullptr;
 	_currentcolor.setCMYK(p[0], p[1], p[2], p[3]);
 	if (_actions)
 		_actions->setColor(_currentcolor);
@@ -1117,7 +1148,7 @@ void PsSpecialHandler::setcmykcolor (vector<double> &p) {
 
 
 void PsSpecialHandler::sethsbcolor (vector<double> &p) {
-	_pattern = 0;
+	_pattern = nullptr;
 	_currentcolor.setHSB(p[0], p[1], p[2]);
 	if (_actions)
 		_actions->setColor(_currentcolor);
@@ -1149,10 +1180,14 @@ void PsSpecialHandler::ClippingStack::pushEmptyPath () {
 
 
 void PsSpecialHandler::ClippingStack::push (const Path &path, int saveID) {
+	shared_ptr<Path> prependedPath;
+	if (!_stack.empty())
+		prependedPath = _stack.top().prependedPath;
 	if (path.empty())
 		_stack.emplace(Entry(saveID));
 	else
 		_stack.emplace(Entry(path, ++_maxID, saveID));
+	_stack.top().prependedPath = prependedPath;
 }
 
 
@@ -1188,21 +1223,20 @@ void PsSpecialHandler::ClippingStack::pop (int saveID, bool grestoreall) {
 
 
 /** Returns a pointer to the path on top of the stack, or 0 if the stack is empty. */
-const PsSpecialHandler::Path* PsSpecialHandler::ClippingStack::top () const {
+const PsSpecialHandler::Path* PsSpecialHandler::ClippingStack::path () const {
 	return _stack.empty() ? nullptr : _stack.top().path.get();
 }
 
 
-/** Returns true if the clipping path was loaded into the graphics path (via PS operator 'clippath') */
-bool PsSpecialHandler::ClippingStack::clippathLoaded () const {
-	return !_stack.empty() && _stack.top().cpathLoaded;
+/** Returns a pointer to the path on top of the stack, or 0 if the stack is empty. */
+const PsSpecialHandler::Path* PsSpecialHandler::ClippingStack::prependedPath () const {
+	return _stack.empty() ? nullptr : _stack.top().prependedPath.get();
 }
 
 
-void PsSpecialHandler::ClippingStack::setClippathLoaded (bool loaded) {
-	if (_stack.empty())
-		return;
-	_stack.top().cpathLoaded = loaded;
+void PsSpecialHandler::ClippingStack::removePrependedPath () {
+	if (!_stack.empty())
+		_stack.top().prependedPath = nullptr;
 }
 
 
@@ -1213,15 +1247,19 @@ void PsSpecialHandler::ClippingStack::clear() {
 }
 
 
-/** Replaces the top element by a new one.
- *  @param[in] path new path to be on top of the stack */
-void PsSpecialHandler::ClippingStack::replace (const Path &path) {
+/** Replaces the top path by a new one.
+ *  @param[in] path new path to put on the stack
+ *  @return true if the new path differs from the previous one */
+bool PsSpecialHandler::ClippingStack::replace (const Path &path) {
 	if (_stack.empty())
 		push(path, -1);
+	else if (_stack.top().path && path == *_stack.top().path)
+		return false;
 	else {
 		_stack.top().path = make_shared<Path>(path);
 		_stack.top().pathID = ++_maxID;
 	}
+	return true;
 }
 
 
@@ -1232,9 +1270,16 @@ void PsSpecialHandler::ClippingStack::dup (int saveID) {
 }
 
 
-const vector<const char*> PsSpecialHandler::prefixes () const {
-	const vector<const char*> pfx {
+void PsSpecialHandler::ClippingStack::setPrependedPath () {
+	if (!_stack.empty())
+		_stack.top().prependedPath = _stack.top().path;
+}
+
+
+vector<const char*> PsSpecialHandler::prefixes() const {
+	vector<const char*> pfx {
 		"header=",    // read and execute PS header file prior to the following PS statements
+		"pdffile=",   // process PDF file
 		"psfile=",    // read and execute PS file
 		"PSfile=",    // dito
 		"ps:",        // execute literal PS code wrapped by @beginspecial and @endspecial
