@@ -2,7 +2,7 @@
 ** PsSpecialHandler.hpp                                                 **
 **                                                                      **
 ** This file is part of dvisvgm -- a fast DVI to SVG converter          **
-** Copyright (C) 2005-2018 Martin Gieseking <martin.gieseking@uos.de>   **
+** Copyright (C) 2005-2019 Martin Gieseking <martin.gieseking@uos.de>   **
 **                                                                      **
 ** This program is free software; you can redistribute it and/or        **
 ** modify it under the terms of the GNU General Public License as       **
@@ -36,7 +36,7 @@
 class PSPattern;
 class XMLElementNode;
 
-class PsSpecialHandler : public SpecialHandler, public DVIEndPageListener, protected PSActions {
+class PsSpecialHandler : public SpecialHandler, protected PSActions {
 	using Path = GraphicsPath<double>;
 	using ColorSpace = Color::ColorSpace;
 
@@ -49,25 +49,26 @@ class PsSpecialHandler : public SpecialHandler, public DVIEndPageListener, prote
 			ClippingStack () : _maxID(0) {}
 			void pushEmptyPath ();
 			void push (const Path &path, int saveID=-1);
-			void replace (const Path &path);
+			bool replace (const Path &path);
 			void dup (int saveID=-1);
 			void pop (int saveID=-1, bool grestore=false);
 			void clear ();
 			bool empty () const {return _stack.empty();}
-			bool clippathLoaded () const;
-			void setClippathLoaded (bool loaded);
-			const Path* top () const;
+			void setPrependedPath ();
+			const Path* path () const;
+			const Path* prependedPath () const;
+			void removePrependedPath ();
 			int topID () const {return _stack.empty() ? 0 : _stack.top().pathID;}
 
 		private:
 			struct Entry {
-				std::shared_ptr<Path> path;  // pointer to current clipping path
+				std::shared_ptr<Path> path;  ///< pointer to current clipping path
+				std::shared_ptr<Path> prependedPath=nullptr;
 				int pathID;        ///< ID of current clipping path
 				int saveID;        ///< if >=0, path was pushed by 'save', and saveID holds the ID of the PS memory object
-				bool cpathLoaded;  ///< true if clipping path was loaded into current path
 				Entry () : Entry(-1) {}
-				Entry (const Path &p, int pid, int sid) : path(std::make_shared<Path>(p)), pathID(pid), saveID(sid), cpathLoaded(false) {}
-				Entry (int sid) : path(nullptr), pathID(0), saveID(sid), cpathLoaded(false) {}
+				Entry (const Path &p, int pid, int sid) : path(std::make_shared<Path>(p)), pathID(pid), saveID(sid) {}
+				Entry (int sid) : path(nullptr), pathID(0), saveID(sid) {}
 				Entry (const Entry &entry) =default;
 				Entry (Entry &&entry) =default;
 			};
@@ -76,17 +77,19 @@ class PsSpecialHandler : public SpecialHandler, public DVIEndPageListener, prote
 	};
 
 	enum PsSection {PS_NONE, PS_HEADERS, PS_BODY};
+	enum class FileType {EPS, PDF};
 
 	public:
 		PsSpecialHandler ();
 		~PsSpecialHandler ();
 		const char* name () const override {return "ps";}
 		const char* info () const override {return "dvips PostScript specials";}
-		const std::vector<const char*> prefixes () const override;
-		void preprocess (const char *prefix, std::istream &is, SpecialActions &actions) override;
-		bool process (const char *prefix, std::istream &is, SpecialActions &actions) override;
+		std::vector<const char*> prefixes() const override;
+		void preprocess (const std::string &prefix, std::istream &is, SpecialActions &actions) override;
+		bool process (const std::string &prefix, std::istream &is, SpecialActions &actions) override;
 		void setDviScaleFactor (double dvi2bp) override {_previewFilter.setDviScaleFactor(dvi2bp);}
 		void enterBodySection ();
+		PSInterpreter& psInterpreter () {return _psi;}
 
 	public:
 		static bool COMPUTE_CLIPPATHS_INTERSECTIONS;
@@ -100,9 +103,9 @@ class PsSpecialHandler : public SpecialHandler, public DVIEndPageListener, prote
 		void moveToDVIPos ();
 		void executeAndSync (std::istream &is, bool updatePos);
 		void processHeaderFile (const char *fname);
-		void psfile (const std::string &fname, const std::unordered_map<std::string,std::string> &attr);
+		void imgfile (FileType type, const std::string &fname, const std::unordered_map<std::string,std::string> &attr);
 		void dviEndPage (unsigned pageno, SpecialActions &actions) override;
-		void clip (Path &path, bool evenodd);
+		void clip (Path path, bool evenodd);
 		void processSequentialPatchMesh (int shadingTypeID, ColorSpace cspace, VectorIterator<double> &it);
 		void processLatticeTriangularPatchMesh (ColorSpace colorSpace, VectorIterator<double> &it);
 
@@ -132,6 +135,7 @@ class PsSpecialHandler : public SpecialHandler, public DVIEndPageListener, prote
 		void rotate (std::vector<double> &p) override;
 		void save (std::vector<double> &p) override;
 		void scale (std::vector<double> &p) override;
+		void setblendmode (std::vector<double> &p) override    {_blendmode = int(p[0]);}
 		void setcmykcolor (std::vector<double> &cmyk) override;
 		void setdash (std::vector<double> &p) override;
 		void setgray (std::vector<double> &p) override;
@@ -142,6 +146,8 @@ class PsSpecialHandler : public SpecialHandler, public DVIEndPageListener, prote
 		void setmatrix (std::vector<double> &p) override;
 		void setmiterlimit (std::vector<double> &p) override   {_miterlimit = p[0];}
 		void setopacityalpha (std::vector<double> &p) override {_opacityalpha = p[0];}
+		void setshapealpha (std::vector<double> &p) override   {_shapealpha = p[0];}
+		void setpagedevice (std::vector<double> &p) override;
 		void setpattern (std::vector<double> &p) override;
 		void setrgbcolor (std::vector<double> &rgb) override;
 		void shfill (std::vector<double> &p) override;
@@ -153,7 +159,7 @@ class PsSpecialHandler : public SpecialHandler, public DVIEndPageListener, prote
 		PSInterpreter _psi;
 		SpecialActions *_actions;
 		PSPreviewFilter _previewFilter;  ///< filter to extract information generated by the preview package
-		PsSection _psSection;       ///< current section processed (nothing yet, headers, or body specials)
+		PsSection _psSection=PS_NONE;    ///< current section processed (nothing yet, headers, or body specials)
 		XMLElementNode *_xmlnode;   ///< if != 0, created SVG elements are appended to this node
 		XMLElementNode *_savenode;  ///< pointer to temporaryly store _xmlnode
 		std::string _headerCode;    ///< collected literal PS header code
@@ -165,6 +171,8 @@ class PsSpecialHandler : public SpecialHandler, public DVIEndPageListener, prote
 		double _linewidth;          ///< current line width in bp units
 		double _miterlimit;         ///< current miter limit in bp units
 		double _opacityalpha;       ///< opacity level (0=fully transparent, ..., 1=opaque)
+		double _shapealpha;         ///< shape opacity level (0=fully transparent, ..., 1=opaque)
+		int _blendmode;             ///< blend mode used when overlaying colored areas
 		uint8_t _linecap  : 2;      ///< current line cap (0=butt, 1=round, 2=projecting square)
 		uint8_t _linejoin : 2;      ///< current line join (0=miter, 1=round, 2=bevel)
 		double _dashoffset;         ///< current dash offset
