@@ -2,7 +2,7 @@
 ** XMLNode.cpp                                                          **
 **                                                                      **
 ** This file is part of dvisvgm -- a fast DVI to SVG converter          **
-** Copyright (C) 2005-2018 Martin Gieseking <martin.gieseking@uos.de>   **
+** Copyright (C) 2005-2019 Martin Gieseking <martin.gieseking@uos.de>   **
 **                                                                      **
 ** This program is free software; you can redistribute it and/or        **
 ** modify it under the terms of the GNU General Public License as       **
@@ -18,6 +18,7 @@
 ** along with this program; if not, see <http://www.gnu.org/licenses/>. **
 *************************************************************************/
 
+#include <algorithm>
 #include <map>
 #include <list>
 #include <sstream>
@@ -148,9 +149,11 @@ bool XMLElementNode::insertAfter (unique_ptr<XMLNode> &&child, XMLNode *sibling)
 
 /** Removes a given child from the element. */
 void XMLElementNode::remove (const XMLNode *child) {
-	_children.remove_if([=](const unique_ptr<XMLNode> &ptr) {
+	auto it = find_if(_children.begin(), _children.end(), [=](const unique_ptr<XMLNode> &ptr) {
 		return ptr.get() == child;
 	});
+	if (it != _children.end())
+		_children.erase(it);
 }
 
 
